
'''
Function to read data from Pandora

Author:  Michel Van Roozendael (Matlab version)
         Caroline FAYT (Python version)
         (Royal Belgian Institute for Space Aeronomy)
         
Creation Date            : September 2020
Modified 13 January 2021 : Determine automatically the indexes of columns and position of filters from the header
Version: v_01.1
'''

import csv;
from doas_lib import *

PANDORA_code=0            # Two letter code of measurement routine (** for manual operation)
PANDORA_date=1            # UT date and time
PANDORA_tott=2            # Total duration of measurement set in seconds (=# if the line is a comment line)
PANDORA_tint=3            # Integration time [ms]
PANDORA_nacc=4            # Number of cycles
PANDORA_pos1=5            # Position of filterwheel #1, 0=filterwheel not used, 1-9 are valid positions
PANDORA_pos2=6            # Position of filterwheel #2, 0=filterwheel not used, 1-9 are valid positions
PANDORA_elev=7            # Pointing zenith angle in degree, absolute or relative (see next column), 999=tracker not used
PANDORA_azim=8            # Pointing azimuth in degree, increases clockwise, absolute (0=north) or relative (see next column), 999=tracker not used
PANDORA_spec=9            # radiance 
PANDORA_max=10

FILTER_drk1=0             # filter 1, OPAQUE position
FILTER_drk2=1             # filter 2, OPAQUE position
FILTER_u340=2             # filter 2, U340 position
FILTER_u340d=3            # filter 2, U340+diffusor position
FILTER_max=4

# description of a PANDORA record

class PANDORA_record(dict) :
    """
    Definition of lines of file (example)

               Column 1: Two letter code of measurement routine (** for manual operation)
               Column 2: UT date and time for beginning of measurement, yyyymmddThhmmssZ (ISO 8601)
               Column 3: Routine count (1 for the first routine of the day, 2 for the second, etc.)
               Column 4: Repetition count (1 for the first set in the routine, 2 for the second, etc.)
               Column 5: Total duration of measurement set in seconds (=# if the line is a comment line)
               Column 6: Integration time [ms]
               Column 7: Number of cycles
               Column 8: Saturation index: positive integer is the number of saturated cycles included in the data, negative integer is the number of cycles skipped due to saturation
               Column 9: Position of filterwheel #1, 0=filterwheel not used, 1-9 are valid positions
               Column 10: Position of filterwheel #2, 0=filterwheel not used, 1-9 are valid positions
               Column 11: Pointing zenith angle in degree, absolute or relative (see next column), 999=tracker not used
               Column 12: Zenith pointing mode: zenith angle is... 0=absolute, 1=relative to sun, 2=relative to moon
               Column 13: Pointing azimuth in degree, increases clockwise, absolute (0=north) or relative (see next column), 999=tracker not used
               Column 14: Azimuth pointing mode: like zenith angle mode but also fixed scattering angles relative to sun (3) or moon (4)
               Column 15: Data processing type index, -9=manual operation
               Column 16: Target distance [m], -1=not pointed on target
               Column 17: Spectrometer control temperature [°C], 999=no temperature signal
               Column 18: Auxiliary spectrometer temperature [°C], 999=no temperature signal
               Column 19: Temperature in head sensor [°C], 999=no temperature signal
               Column 20: Temperature at electronics board 1 [°C], 999=no temperature signal
               Column 21: Temperature at electronics board 2 [°C], 999=no temperature signal
               Column 22: Humidity in head sensor [%], -9=no humidity signal
               Column 23: Pressure in head sensor [hPa], -9=no pressure signal
               Column 24: Scale factor for data (to obtain unscaled data and uncertainty divide then by this number)
               Column 25: Uncertainty indicator: uncertainty is... 0=not given, 1=standard deviation, 2=rms to a fitted straight line
               Columns 26-2077: Mean over all cycles of raw counts for each pixel
               Columns 2078-4129: Uncertainty of raw counts for each pixel divided by the square root of the number of cycles

    Filter positions (example)

               Filterwheel 1, position 1 -> OPEN
               Filterwheel 1, position 2 -> OPEN
               Filterwheel 1, position 3 -> ND0.5
               Filterwheel 1, position 4 -> OPEN
               Filterwheel 1, position 5 -> ND0.9
               Filterwheel 1, position 6 -> ND1.3
               Filterwheel 1, position 7 -> ND1.3
               Filterwheel 1, position 8 -> BP300
               Filterwheel 1, position 9 -> OPAQUE

               Filterwheel 2, position 1 -> OPEN
               Filterwheel 2, position 2 -> DIFF
               Filterwheel 2, position 3 -> OPAQUE
               Filterwheel 2, position 4 -> OPEN
               Filterwheel 2, position 5 -> DIFF
               Filterwheel 2, position 6 -> ND4
               Filterwheel 2, position 7 -> U340
               Filterwheel 2, position 8 -> U340+DIFF
               Filterwheel 2, position 9 -> ND4
    """
    def __init__(self):
       dict();

    def init(self):

       self["r"]=[];
       self["total_tint"]=[];
       self["tint"]=[];
       self["nacc"]=[];
       self["pos1"]=[];
       self["pos2"]=[];
       self["elev"]=[];
       self["azim"]=[];
       self["rad"]=[];
       self["radcor"]=[];
       self["ymdhms"]=[];
       self["fday"]=[];
       self["ftime"]=[];
       self["mtype"]=[];
       self["tpos2"]=[];
       self["sza"]=[];
       self["saa"]=[];

    def parse(self,fields_list,m_type,longit,latit,npix,fieldsIndex,filtersIndex):
     
       self["r"]=fields_list[fieldsIndex[PANDORA_code]];                                                                    # Two letter code of measurement routine (** for manual operation)
       if len(fields_list[fieldsIndex[PANDORA_date]])==16:
        dt=datetime.strptime(fields_list[fieldsIndex[PANDORA_date]],'%Y%m%dT%H%M%SZ');                                      # UT date and time for beginning of measurement, yyyymmddThhmmssZ (ISO 8601)
        ms=0;
       else: 
        dt=datetime.strptime(fields_list[fieldsIndex[PANDORA_date]],'%Y%m%dT%H%M%S.%fZ');                                   # UT date and time for beginning of measurement, yyyymmddThhmmss.fZ (ISO 8601) with f, thenths of seconds
        ms=round(dt.microsecond/1000);
        
       self["total_tint"]=double(fields_list[fieldsIndex[PANDORA_tott]]);                                                   # Total duration of measurement set in seconds (=# if the line is a comment line)
       
       self["tint"]=double(fields_list[fieldsIndex[PANDORA_tint]])*0.001;                                                   # Integration time [ms]
       self["nacc"]=int(fields_list[fieldsIndex[PANDORA_nacc]]);                                                            # Number of cycles
       self["pos1"]=int(fields_list[fieldsIndex[PANDORA_pos1]]);                                                            # Position of filterwheel #1, 0=filterwheel not used, 1-9 are valid positions
       self["pos2"]=int(fields_list[fieldsIndex[PANDORA_pos2]]);                                                            # Position of filterwheel #2, 0=filterwheel not used, 1-9 are valid positions
       self["elev"]=90.-double(fields_list[fieldsIndex[PANDORA_elev]]);                                                     # Pointing zenith angle in degree, absolute or relative (see next column), 999=tracker not used
       self["azim"]=double(fields_list[fieldsIndex[PANDORA_azim]]);                                                         # Pointing azimuth in degree, increases clockwise, absolute (0=north) or relative (see next column), 999=tracker not used
       self["rad"]=asarray([double(f) for f in fields_list[fieldsIndex[PANDORA_spec]:fieldsIndex[PANDORA_spec]+npix]]);     # Mean over all cycles of raw counts for each pixel
       self["radcor"]=[];                                                                                                   # radiance corrected with dark current

       self["ymdhms"]=asarray([dt.timetuple().tm_year,dt.timetuple().tm_mon,dt.timetuple().tm_mday, \
                            dt.timetuple().tm_hour,dt.timetuple().tm_min,dt.timetuple().tm_sec],dtype=int16);
       
       self["fday"]=dt.timetuple().tm_yday;                                                                                 # day of year
       self["ftime"]=self["ymdhms"][3]+self["ymdhms"][4]/60+self["ymdhms"][5]/3600;                                         # fractional time
       
       if self["pos1"]==filtersIndex[FILTER_drk1] or self["pos2"]==filtersIndex[FILTER_drk2]:
          self["mtype"]=QDOAS_MeasType_dark;
       elif self["r"][0]=='E':
          self["mtype"]=QDOAS_MeasType_offaxis if self["elev"]<80. else QDOAS_MeasType_zenith;
       elif self["r"][0]=='S':
          self["mtype"]=QDOAS_MeasType_ds;
       elif self["r"][0]=='Z':
          self["mtype"]=QDOAS_MeasType_zenith;
       else:
          self["mtype"]=0;
          
       if self["mtype"]==0:
          self["tpos2"]=False;
       elif m_type==1 and (self["pos2"]!=filtersIndex[FILTER_drk2]) and ((self["pos2"]==filtersIndex[FILTER_u340]) or (self["pos2"]==filtersIndex[FILTER_u340d])): # UV
          self["tpos2"]=True;
       elif ((m_type==2) or (m_type==3)) and (self["pos2"]!=filtersIndex[FILTER_drk2]) and ((self["pos2"]!=filtersIndex[FILTER_u340]) and (self["pos2"]!=filtersIndex[FILTER_u340d])): # UVVIS
          self["tpos2"]=True;
       else:
          self["tpos2"]=False;
       
       (self["sza"],self["saa"])=sunangl(self["ymdhms"][0:3],[self["ymdhms"][3:6]],-longit,latit);   # !!! use brackets for ut to have dimension 2 -> see sunangl
       self["saa"]=180.-self["saa"];

       # !!! to check if it's necessary with netcdf output self["saa"]=self["saa"]-180.;

    def append(self,record):

       if len(self)==0:
          self.init();

       self["r"].append(record["r"]);
       self["total_tint"].append(record["total_tint"]);
       self["tint"].append(record["tint"]);
       self["nacc"].append(record["nacc"]);
       self["pos1"].append(record["pos1"]);
       self["pos2"].append(record["pos2"]);
       self["elev"].append(record["elev"]);
       self["azim"].append(record["azim"]);
       self["rad"].append(record["rad"]);
       self["radcor"].append(record["rad"]);
       self["ymdhms"].append(record["ymdhms"]);
       self["fday"].append(record["fday"]);
       self["ftime"].append(record["ftime"]);
       self["mtype"].append(record["mtype"]);
       self["tpos2"].append(record["tpos2"]);
       self["sza"].append(record["sza"]);
       self["saa"].append(record["saa"]);


def extract_pandora(sp_file,of_file,m_type,longitude,latitude,file_type="L1"):
    """
    return the array of Pandora records

    Parameters
    ----------

    sp_file        : the name of the radiance file
    of_file        : the name of the instrument operation file
    m_type         : defines the band

         m_type=1 -> UV (bandpass filter U340 in place, s1 files)
         m_type=2 -> UVVIS (complete spectra from s1 files)
         m_type=3 -> VIS-NIR (400-900) to apply on s2 siles
         
    longitude      : the longitude of the observation site
    latitude       : the latitude of the observation site
    file_type      : the level of the radiance spectra (L0 or L1); default is L1

    Returns
    -------

    wl          : the wavelength calibration constructed with the dispersion polynomial
    spe_records : the array of radiances records (include original spectra and spectra corrected by dark currents)
    drk_records : the array of dark current
    """

    functionName="extract_pandora"

    spe_records=PANDORA_record();
    drk_records=PANDORA_record();
    rec=PANDORA_record();
    fieldsIndex=-ones(PANDORA_max,dtype=int);
    filtersIndex=zeros(FILTER_max,dtype=int);

    if not os.path.isfile(sp_file):
       print('['+functionName+'] '+sp_file+' not found !!!');
    elif file_type=="L0" and not os.path.isfile(of_file):
       print('['+functionName+'] '+of_file+' not found !!!');
    elif m_type!=1 and m_type!=2 and m_type!=3:
       print('['+functionName+'] m_type should be 1 (UV), 2 (UVVIS) or 3 (VIS-NIR)');
    else:
     
       # retrieve parameters from of file
       
       print('Read '+of_file);

       of_dict={};

       with open(of_file) as fd:
            reader=csv.reader(fd)
            for line in reader:
                if len(line)>1:
                   line=[','.join(line)];   # avoid that lines with a comma are splitted
                if len(line)>0:
                   key=line[0].split(" -> ")[0]
                   value=line[0].split(" -> ")[1];
                   of_dict[key]=value;
                   
                   filterPosition=int(key.split(' ')[-1]) if key.startswith('Filterwheel') else 0;
                      
                   if key.startswith('Filterwheel 1, position') and value=="OPAQUE":
                      filtersIndex[FILTER_drk1]=filterPosition;    
                   elif key.startswith('Filterwheel 2, position') and value=="OPAQUE":     
                      filtersIndex[FILTER_drk2]=filterPosition;
                   elif key.startswith('Filterwheel 2, position') and value=="U340":     
                      filtersIndex[FILTER_u340]=filterPosition;           
                   elif key.startswith('Filterwheel 2, position') and value=="U340+DIFF":     
                      filtersIndex[FILTER_u340d]=filterPosition;           

            npix=asarray(list(map(int,of_dict["Number of pixels"].split(' '))))[0];
            
       print(filtersIndex[FILTER_drk1],filtersIndex[FILTER_drk2],filtersIndex[FILTER_u340],filtersIndex[FILTER_u340d]);
       
       if file_type == "L0":
        
          if m_type==1 or m_type==2:
             p=asarray(list(map(double,of_dict["Dispersion polynomial"].split(' '))))
             dc_offset=double(of_dict["Dark offset [counts]"]);
             dc_slope=double(of_dict["Dark slope [counts/s]"]);
          else:
             p=asarray(list(map(double,of_dict["Dispersion polynomial 2"].split(' '))))
             dc_offset=double(of_dict["Dark offset [counts] 2"]);
             dc_slope=double(of_dict["Dark slope [counts/s] 2"]);
             
          # wavelength calibration

          x=arange(1,npix+1);                       
          xs=3.46*(x/npix-0.5);                    # scaled pixels
          wl=polyval(p,xs);                        # wavelength scale
          
       else:
          npix=2048;
          
       # open file, encoding allows special character like degree, percent...

       print('Read '+sp_file);
       
       headerFlag=0;
       
       # Read the file line by line
       
       with open(sp_file,'r',encoding='ISO-8859-1') as fd: 
         reader=csv.reader(fd); 
         for line in reader: 
          
             if headerFlag==2:
              
                if len(line)>1:
                   line=[','.join(line)];   # avoid that lines with a comma are splitted
                   
                fields_list=line[0].split(' ');
                
                if fields_list[4]!='#':
                   rec.parse(fields_list,m_type,longitude,latitude,npix,fieldsIndex,filtersIndex);
                   if rec["mtype"]==QDOAS_MeasType_dark:
                      drk_records.append(rec);
                   elif rec["pos1"]!=filtersIndex[FILTER_drk1] and rec["tpos2"]==True:
                      spe_records.append(rec);
                      
             # header with the description of columns
             
             elif line[0].startswith('---'):   
                headerFlag=headerFlag+1;
             elif line[0].startswith('Column '): 
                
                columnIndex=int(line[0].split(':')[0].split(' ')[1])-1; 
                columnTitle=line[0].split(':')[1].lstrip();
                pandoraIndex=-1;
                
                if columnTitle.startswith('Two letter code'):
                   pandoraIndex=PANDORA_code;
                elif columnTitle.startswith('UT date and time'):
                   pandoraIndex=PANDORA_date;
                elif columnTitle.startswith('Total duration of measurement'):
                   pandoraIndex=PANDORA_tott;
                elif columnTitle.startswith('Integration time [ms]'):
                   pandoraIndex=PANDORA_tint;
                elif columnTitle.startswith('Number of cycles') or columnTitle.startswith('Number of bright count cycles'):
                   pandoraIndex=PANDORA_nacc;
                elif columnTitle.startswith('Position of filterwheel #1') or columnTitle.startswith('Effective position of filterwheel #1') :
                   pandoraIndex=PANDORA_pos1;
                elif columnTitle.startswith('Position of filterwheel #2') or columnTitle.startswith('Effective position of filterwheel #2') :
                   pandoraIndex=PANDORA_pos2;
                elif columnTitle.startswith('Pointing zenith angle in degree'):
                   pandoraIndex=PANDORA_elev;
                elif columnTitle.startswith('Pointing azimuth in degree'):
                   pandoraIndex=PANDORA_azim;
                 
                if pandoraIndex!=-1:
                   fieldsIndex[pandoraIndex]=columnIndex;
                   
             elif line[0].startswith('Columns '):
              
                  if ((file_type=="L0") and line[0].split(':')[1].lstrip().startswith('Mean over all cycles')) or ((file_type=="L1") and line[0].split(':')[1].lstrip().startswith('Level 1 data for each pixel')): 
                   fieldsIndex[PANDORA_spec]=int(line[0].split(':')[0].split(' ')[1].split('-')[0])-1;
                     
             elif line[0].startswith('Nominal wavelengths [nm]:') and file_type=="L1":
                wl=asarray([double(n) for n in line[0].split(':')[1].lstrip().split(' ')])

       nspe=len(spe_records["tint"]);   
       
       # dark current correction
       
       if file_type=="L0":
       
          for ispe in arange(0,nspe):
           
              spe_tint=spe_records["tint"][ispe];
              spe_ftime=spe_records["ftime"][ispe];
              spe_nacc=spe_records["nacc"][ispe];
              
              i=[i for i,(drk_tint,drk_ftime) in enumerate(zip(drk_records["tint"],drk_records["ftime"])) if abs(drk_tint-spe_tint)<0.05 and abs(drk_ftime-spe_ftime)<0.25 ];
              if len(i)>0:
                 j=[j for j,drk_ftime in enumerate(asarray(drk_records["ftime"])[i]) if abs(drk_ftime-spe_ftime)==min(abs(asarray(drk_records["ftime"])[i]-spe_ftime))]
                 idrk=i[j[0]];
                 drk=drk_records["rad"][idrk]*spe_records["nacc"][ispe]/drk_records["nacc"][idrk]
                 spe=spe_records["rad"][ispe];
                 spe_records["rad"][ispe]=spe[0:2048];
                 spe=spe[0:2048]-drk[0:2048];
                 spe_records["radcor"][ispe]=spe;
              else:
                 print('record '+i+' not dark current corrected');
                 
       else:
          for ispe in arange(0,nspe):
              spe_records["radcor"][ispe]=spe_records["rad"][ispe];
              
       wl=wl[0:2048];
       
    return wl,spe_records;
